PImage bg;
ParticleSystem ps, boom;
ArrayList<Rocket> rockets;
float gravity = 0.05; 
int density = 15;
float boomsize = random(0,5000);
float launchx, launchy;

void setup() {
//  size(displayWidth,displayHeight-40);
  size(1000,597);
  bg = loadImage("austin_night_skyline.jpg");
  launchx = width/2;
  launchy = height-120;
  rockets = new ArrayList<Rocket>();
  ps = new ParticleSystem(new PVector(launchx,launchy), 60, 1, 1);
  boom = new ParticleSystem(new PVector(width/2,100), 10, 5, 1);
  noStroke();
  colorMode(HSB,100);
}

void mousePressed() {
  newBoom(mouseX,mouseY);
}
  
void draw() {
//  background(0);
  background(bg);
  boom.run();
  for (int i = 0; i < density; ++i) ps.addParticle();
  ps.run();
  if (random(1,1000) < 20)
    rockets.add(new Rocket(new PVector(launchx,launchy),
      new PVector(random(-0.05,0.05),random(-.325,-.25)),60,23,1,1));
  for (int i = rockets.size()-1; i >= 0; i--) {
    Rocket r = rockets.get(i);
    r.run();
    if (r.isDead()) {
      newBoom(r.loc.x,r.loc.y);
      rockets.remove(i);
    }
  }
}

void newBoom(float x, float y) {
  boomsize = random(0,5000);
  boom.size = boomsize/1000;
  boom.origin.x = x;
  boom.origin.y = y;
  boom.tint = random(0,100);
  boom.fillParticles(int(boomsize));
}

class ParticleSystem {
  ArrayList<Particle> particles;
  PVector origin;
  float tint;
  float size;
  float decay;

  ParticleSystem(PVector location, float hue, float s, float d) {
    origin = location.get();
    particles = new ArrayList<Particle>();
    tint = hue;
    size = s;
    decay = d;
  }
  
  void fillParticles(int num) {
    for (int i = 0; i < num; i++) addParticle();
  }

  void addParticle() {
    particles.add(new Particle(origin, new PVector(0,0), tint, size, decay));
  }
  
  void addParticle(PVector loc, PVector dir, float c, float s, float d) {
    particles.add(new Particle(loc, dir, c, s, d));
  }

  void run() {
    for (int i = particles.size()-1; i >= 0; i--) {
      Particle p = particles.get(i);
      p.run();
      if (p.isDead()) particles.remove(i);
    }
  }
}


class Particle {
  PVector location;
  PVector velocity;
  PVector acceleration;
  float lifespan;
  float radius;
  float sh,ss,sb,dh,ds,db;
  color c;
  float decay;

  Particle(PVector l, PVector dir, float hue, float pow, float d) {
    decay = d;
    radius = random(1,4);
    lifespan = random(0,25) * radius;
    acceleration = new PVector(0,gravity);
    if (dir.x == 0 && dir.y == 0) {
      velocity = new PVector(random(-1,1),random(-1,1));
      velocity.normalize();
    } else velocity = dir;
    velocity.mult(sqrt(random(0,0.75*pow)));
    location = l.get();
    sh = hue;
    sb = 100;
    ss = 0;
    dh = -sh/lifespan*1.25;
    ds = 100/lifespan;
    ds = ds*2;
    db = 0;
    c = color(sh, ss, sb);
  }

  void run() {
    update();
    display();
  }

  void update() {
    if (location.y > launchy) lifespan = 0;
    else {
    velocity.add(acceleration);
    location.add(velocity);
    sh += dh;
    ss += ds;
    sb += db;
    c = color(sh,ss,sb);
    lifespan -= decay;
    }
  }

  void display() {
    fill(c,lifespan);
    ellipse(location.x,location.y,radius,radius);
  }
  
  boolean isDead() {
    if (lifespan <= 0.0) return true;
    else return false;
  }
}

class Rocket {
  PVector origin;
  PVector loc;
  PVector vel,opp;
  PVector accel;
  float impulse;
  float tint;
  int decay;
  float power;
  int lifetime;
  ParticleSystem exhaust;
  
  Rocket(PVector o, PVector d, float col, float i, float p, int l) {
    exhaust = new ParticleSystem(o,col,p,l);
    reset(o,d,col,i,p,l);
  }
 
  void reset(PVector o, PVector d, float col, float i, float p, int l) {
    origin = loc = o;
    accel = d;
    accel.y += gravity;
    vel = new PVector(0,0);
    opp = new PVector(0,0);
    impulse = i;
    tint = col;
    power = p;
    decay = l;
    lifetime = int(random(0,40));
  }
  
  boolean isDead() {
    if (lifetime < 0) return true;
    else return false;
  }
  
  void run() {
    float density = 10;
    if (impulse <= 0) {
      accel.x = 0;
      accel.y = gravity;
      density = 2;
    } else impulse--;
    if (vel.y > 0) lifetime--;
    opp.x = -vel.x;
    opp.y = -vel.y;
    for (int i = 0; i < density; ++i)
      exhaust.addParticle(loc,new PVector(-vel.x,-vel.y),tint,0.1,1);
    exhaust.run();
    vel.add(accel);
    loc.add(vel);
  }
}

